@@ -30,16 +30,17 @@
 
 #include <libnautilus-extension/nautilus-menu-provider.h>
 
+#include <glib.h>
 #include <glib/gi18n-lib.h>
 #include <gio/gio.h>
 #include <gtk/gtk.h>
 #include <gconf/gconf.h>
 #include <gconf/gconf-client.h>
 #include <libgnome/gnome-desktop-item.h>
-#include <libgnomevfs/gnome-vfs.h>
 
 #include <errno.h>
 #include <fcntl.h>
+#include <stdlib.h> /* for atoi */
 #include <string.h> /* for strcmp */
 #include <unistd.h> /* for chdir */
 #include <sys/stat.h>
@@ -114,31 +115,87 @@ desktop_is_home_dir ()
 				      NULL);
 }
 
+/* a very simple URI parsing routine from Launchpad #333462, until GLib supports URI parsing (GNOME #489862) */
+#define SFTP_PREFIX "sftp://"
+static void
+parse_sftp_uri (GFile *file,
+		char **user,
+		char **host,
+		unsigned int *port,
+		char **path)
+{
+	char *tmp, *save;
+	char *uri;
+
+	uri = g_file_get_uri (file);
+	g_assert (uri != NULL);
+	save = uri;
+
+	*path = NULL;
+	*user = NULL;
+	*host = NULL;
+	*port = 0;
+
+	/* skip intial 'sftp:// prefix */
+	g_assert (!strncmp (uri, SFTP_PREFIX, strlen (SFTP_PREFIX)));
+	uri += strlen (SFTP_PREFIX);
+
+	/* cut out the path */
+	tmp = strchr (uri, '/');
+	if (tmp != NULL) {
+		*path = g_uri_unescape_string (tmp, "/");
+		*tmp = '\0';
+	}
+
+	/* read the username - it ends with @ */
+	tmp = strchr (uri, '@');
+	if (tmp != NULL) {
+		*tmp++ = '\0';
+
+		*user = strdup (uri);
+		if (strchr (*user, ':') != NULL) {
+			/* chop the password */
+			*(strchr (*user, ':')) = '\0'; 
+		}
+
+		uri = tmp;
+	}
+
+	/* now read the port, starts with : */
+	tmp = strchr (uri, ':');
+	if (tmp != NULL) {
+		*tmp++ = '\0';
+		*port = atoi (tmp);  /*FIXME: getservbyname*/
+	}
+
+	/* what is left is the host */
+	*host = strdup (uri);
+	g_free (save);
+}
+
 static char *
 get_remote_ssh_command (const char *uri,
 			const char *command_to_run)
 {
-	GnomeVFSURI *vfs_uri;
-	const char *host_name, *path, *user_name;
+	GFile *file;
+
+	char *host_name, *path, *user_name;
 	char *command, *user_host, *unescaped_path;
+	char *quoted_path;
+	char *remote_command;
+	char *quoted_remote_command;
 	char *port_str;
 	guint host_port;
 
 	g_assert (uri != NULL);
-	g_assert (strncmp (uri, "sftp", strlen ("sftp")) == 0 ||
-		  strncmp (uri, "ssh", strlen ("ssh")) == 0);
 
-	gnome_vfs_init ();
+	file = g_file_new_for_uri (uri);
+	parse_sftp_uri (file, &user_name, &host_name, &host_port, &path);
+	g_object_unref (file);
 
-	vfs_uri = gnome_vfs_uri_new (uri);
-	g_assert (vfs_uri != NULL);
-
-	host_name = gnome_vfs_uri_get_host_name (vfs_uri);
-	host_port = gnome_vfs_uri_get_host_port (vfs_uri);
-	user_name = gnome_vfs_uri_get_user_name (vfs_uri);
-	path = gnome_vfs_uri_get_path (vfs_uri);
 	/* FIXME to we have to consider the remote file encoding? */
-	unescaped_path = gnome_vfs_unescape_string (path, NULL);
+	unescaped_path = g_uri_unescape_string (path, NULL);
+	quoted_path = g_shell_quote (unescaped_path);
 
 	port_str = NULL;
 	if (host_port != 0) {
@@ -154,15 +211,26 @@ get_remote_ssh_command (const char *uri,
 	}
 
 	if (command_to_run != NULL) {
-		command = g_strdup_printf ("ssh %s%s -t \"cd \'%s\' && %s\"", user_host, port_str, unescaped_path, command_to_run);
+		remote_command = g_strdup_printf ("cd %s && exec %s", quoted_path, command_to_run);
 	} else {
-		command = g_strdup_printf ("ssh %s%s -t \"cd \'%s\' && $SHELL -l\"", user_host, port_str, unescaped_path);
+		remote_command = g_strdup_printf ("cd %s && exec $SHELL -l", quoted_path);
 	}
 
+	quoted_remote_command = g_shell_quote (remote_command);
+
+	command = g_strdup_printf ("ssh %s -t %s", user_host, quoted_remote_command);
+
+	g_free (user_name);
 	g_free (user_host);
-	g_free (unescaped_path);
+	g_free (host_name);
 	g_free (port_str);
-	gnome_vfs_uri_unref (vfs_uri);
+
+	g_free (path);
+	g_free (unescaped_path);
+	g_free (quoted_path);
+
+	g_free (remote_command);
+	g_free (quoted_remote_command);
 
 	return command;
 }
@@ -185,7 +253,7 @@ get_terminal_command_for_file_info (NautilusFileInfo *file_info,
 				    const char *command_to_run,
 				    gboolean remote_terminal)
 {
-	char *uri, *path;
+	char *uri, *path, *quoted_path;
 	char *command;
 
 	uri = nautilus_file_info_get_activation_uri (file_info);
@@ -206,7 +274,6 @@ get_terminal_command_for_file_info (NautilusFileInfo *file_info,
 			} else {
 				path = g_strdup (g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP));
 			}
-			g_free (path);
 			break;
 
 		case FILE_INFO_SFTP:
@@ -228,11 +295,15 @@ get_terminal_command_for_file_info (NautilusFileInfo *file_info,
 	}
 
 	if (command == NULL && path != NULL) {
+		quoted_path = g_shell_quote (path);
+
 		if (command_to_run != NULL) {
-			command = g_strdup_printf ("cd '%s' && %s", path, command_to_run);
+			command = g_strdup_printf ("cd %s && exec %s", quoted_path, command_to_run);
 		} else {
-			command = g_strdup_printf ("cd '%s' && $SHELL -l", path);
+			command = g_strdup_printf ("cd %s && exec $SHELL -l", quoted_path);
 		}
+
+		g_free (quoted_path);
 	}
 
 	g_free (path);
@@ -395,6 +466,25 @@ terminal_locked_down (void)
                                       NULL);
 }
 
+/* used to determine for remote URIs whether GVFS is capable of mapping them to ~/.gvfs */
+static gboolean
+uri_has_local_path (const char *uri)
+{
+	GFile *file;
+	char *path;
+	gboolean ret;
+
+	file = g_file_new_for_uri (uri);
+	path = g_file_get_path (file);
+
+	ret = (path != NULL);
+
+	g_free (path);
+	g_object_unref (file);
+
+	return ret;
+}
+
 static GList *
 nautilus_open_terminal_get_background_items (NautilusMenuProvider *provider,
 					     GtkWidget		  *window,
@@ -413,24 +503,28 @@ nautilus_open_terminal_get_background_items (NautilusMenuProvider *provider,
 
 	uri = nautilus_file_info_get_activation_uri (file_info);
 	terminal_file_info = get_terminal_file_info (uri);
-	g_free (uri);
 
-	item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
-					    NULL, terminal_file_info == FILE_INFO_SFTP, FALSE);
-	items = g_list_append (items, item);
+	if (terminal_file_info == FILE_INFO_SFTP || uri_has_local_path (uri)) {
+		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
+						    NULL, terminal_file_info == FILE_INFO_SFTP, FALSE);
+		items = g_list_append (items, item);
+	}
 
-	if (terminal_file_info == FILE_INFO_SFTP) {
+	if (terminal_file_info == FILE_INFO_SFTP && uri_has_local_path (uri)) {
 		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
 						    NULL, FALSE, FALSE);
 		items = g_list_append (items, item);
 	}
 
 	if (display_mc_item () &&
-	    g_find_program_in_path ("mc")) {
+	    g_find_program_in_path ("mc") &&
+	    uri_has_local_path (uri)) {
 		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window), "mc", FALSE, FALSE);
 		items = g_list_append (items, item);
 	}
 
+	g_free (uri);
+
 	return items;
 }
 
@@ -459,23 +553,26 @@ nautilus_open_terminal_get_file_items (NautilusMenuProvider *provider,
 
 	uri = nautilus_file_info_get_activation_uri (files->data);
 	terminal_file_info = get_terminal_file_info (uri);
-	g_free (uri);
 
 	switch (terminal_file_info) {
 		case FILE_INFO_LOCAL:
 		case FILE_INFO_SFTP:
 		case FILE_INFO_OTHER:
-			item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window),
-							    NULL, terminal_file_info == FILE_INFO_SFTP, TRUE);
-			items = g_list_append (items, item);
+			if (terminal_file_info == FILE_INFO_SFTP || uri_has_local_path (uri)) {
+				item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window),
+								    NULL, terminal_file_info == FILE_INFO_SFTP, TRUE);
+				items = g_list_append (items, item);
+			}
 
-			if (terminal_file_info == FILE_INFO_SFTP) {
+			if (terminal_file_info == FILE_INFO_SFTP &&
+			    uri_has_local_path (uri)) {
 				item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window), NULL, FALSE, TRUE);
 				items = g_list_append (items, item);
 			}
 
 			if (display_mc_item () &&
-			    g_find_program_in_path ("mc")) {
+			    g_find_program_in_path ("mc") &&
+			     uri_has_local_path (uri)) {
 				item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window), "mc", TRUE, FALSE);
 				items = g_list_append (items, item);
 			}
@@ -488,6 +585,8 @@ nautilus_open_terminal_get_file_items (NautilusMenuProvider *provider,
 			g_assert_not_reached ();
 	}
 
+	g_free (uri);
+
 	return items;
 }
 
